from pymatgen.symmetry.analyzer import SpacegroupAnalyzer as SGA

def get_spacegroup_info(struct):
    sga = SGA(struct)
    return sga.get_space_group_symbol(), sga.get_space_group_number(), sga.get_crystal_system()


from glob import glob
from yamet.common.utils import *
from yamet.workflows.vasp.getters import get_pressure
from yamet.workflows.vasp.handlers import handle_hubbard, handle_magmoms
from yamet.analysis.symmetry import *
from copy import copy, deepcopy
from ase.calculators.vasp import Vasp
from ase.io import read as ase_read
from ase.core import Atoms
from xml.etree.ElementTree import ParseError
import shutil
from typing import List, Tuple, Any, Dict, Union
import os
from pymatgen.io.vasp.outputs import Vasprun, BSVasprun

#TODO:
# * Check dirs for unfinished calcs at start before loop
# * Check for keyword in INCAR, like KSPACING, LWAVE etc.
# * TESTING!

class Workflow:
    def __init__(self, inputs: Union[List[str], str], incars: List[str], names: List[str], kpoints: Union[None, List[str]] = None, HPC_mode = True) -> None:
        
        self.globbed_inputs = []
        self.inp_type = type(inputs)
        self.wf_dirs = ["inputs", "convered", "misc"]
        if isinstance(inputs, List):
            for inp in inputs:
                self.globbed_inputs += glob(inp)
        elif isinstance(inputs, str):
            self.globbed_inputs = glob(inputs)
        else:
            raise TypeError("Inputs must be given as a list or wildcarded (with '*')\nstring of the file locations")
        
        for d in self.wf_dirs:
            os.mkdir(d)
        # As the function says
        self.ensure_exist(incars)
        if kpoints is not None:
            self.ensure_exist(kpoints)
            
        self.init_magmoms = []
        self.hubbard = []
            
        # Also make sure that incars contain lines to print WAVECAR and CHGCAR for mult-step runs, and kpoints are present if kspacing not defined.
        for incar in incars:
            pass
        self.incars = incars
        self.kpoints = kpoints
        
        
    def run_vasp(self, atoms: Atoms, incar: str, restart: bool = False) -> Atoms:
        calc = Vasp(restart = restart, txt = "vasp.out")
        calc.set(ldua_luj = handle_hubbard(c_atoms, self.hubbard))
        if restart:
            c_atoms = calc.get_atoms()
        else:
            c_atoms = atoms
            calc.read_incar(filename = incar)
            calc.set(setups = "recommended")
            c_atoms.set_calculator(calc)
            c_atoms = handle_magmoms(c_atoms, self.init_magmoms)
        
        # Running the damn calculation
        energy = atoms.get_potential_energy()
        return atoms
    
    
    def post_vasp(self, name: str, solo = True, continue_if_not_converged: bool = True, next_calc: Any = None, index:int = 0) -> Any:
        readable, vasprun = self.read_vasprun_xml()
        if readable:
            atoms = ase_read("vasprun.xml")
        else:
            print(f"ERROR: {name} {index} need attention - unreadble vasprun.xml")
            return 1, None            
        if vasprun.converged_ionic:
            if not vasprun.converged_electronic:
                print(f"WARN: Final electronic step in {name} run {index} not converged.")
            return 0, atoms
        else:
            print(f"WARN: Geomtery not converged in {name} run {index}")
            if solo or not continue_if_not_converged:
                return 2, atoms
            elif continue_if_not_converged:
                return 0, atoms
                
                
                               
    
    def input_iteration(self, inputf: str, ignore_intermediate_warnings: bool = False, remove_wavecar: bool = False) -> bool:
        """The function to be used for each input in the main running loop.

        Args:
            inputf (str): The path of the input geometry.
            ignore_intermediate_warnings (bool, optional): Defaults to False.
            remove_wavecar (bool): Delete the WAVECAR from the previous calculation. Defaults to False

        Returns:
            bool: Whether the final structure was converged.
        """
        name = ".".join(inputf.split("/")[0].split(".")[:-1])
        skip = False
        with cd(name):
            pwd = os.getcwd()
            shutil.copy2(f"../{inputf}", pwd)
            shutil.move(f"../{inputf}", "../inputs")
            atoms = ase_read(inputf)
            if len(self.incars) == 1:
                shutil.copy2(f"../{incar[0]}", ".")
                out_atoms = self.run_vasp(atoms, incar=self.incars[0])
                
                # Finishing up
                exit_code = self.post_vasp(name = name)
                    
                
            else:
                for idx, incar in enumerate(self.incars):
                    with cd(incar):
                        if not skip:
                            shutil.copy2(f"../../{incar}", ".")
                            if idx > 0:
                                shutil.copy2(f"../{self.incars[idx-1]}/CONTCAR", "./POSCAR")
                                
                                """
                                We need the WAVECAR to carry over the information regarding the magnetic moments
                                determined from the previous run. It will also help with starting the 
                                calculations as VASP won't be beginning from nothing.
                                """
                                shutil.copy2(f"../{self.incars[idx-1]}/WAVECAR", ".")
                            try:
                                shutil.copy2(f"../../{self.kpoints[idx]}", ".")
                            except FileNotFoundError:
                                print(f"INFO: KPOINTS not found for run {idx} in {name}")
                            
                            out_atoms = self.run_vasp(atoms, incar=self.incars)
                            exit_code, atoms = self.post_vasp(name=name, index=idx, solo=False, continue_if_not_converged=True)
                            if exit_code != 0:
                                skip = True # Since breaking inside the with cd() won't return to parent dir...
                            if len(self.incars) == idx + 1:
                                vasprun = Vasprun("vasprun.xml")
                                struct = atoms2struct(atoms)
                                atoms2res(atoms=atoms, fname=f"{name}_final.res", pressure=get_pressure(), energy=vasprun.final_energy, spacegroup=get_spacegroup_info(struct=struct))
                                shutil(f"{name}_final.res", "../../converged")
                        else:
                            pass

            
    def check_dirs(self):
        print("Checking directories for unfinished runs.")
        dirs = glob("*/")
        
        
    def read_vasprun_xml(self, fname: str = "vasprun.xml"):
        assert os.path.exists(f"{os.getcwd()}/{fname}")
        v = None
        try:
            v = Vasprun(fname)
            success = True
        except ParseError:
            print("XML could not be read. Likely that the job scheduler killed the run before VASP could terminate cleanly/successfully. Consider adding a flag to you script to generate a STOPCAR before the job scheduler kills the run.")
            success = False
        except:
            print("Well I din't expect this error. What have you done?")
            success = False
        return success, v
     
    
    def ensure_exist(self, files):
        for f in files:
            if not os.path.exists(f"{os.getcwd()}/{f}"):
                if not os.path.exists(f):
                    raise FileNotFoundError(f"File '{f}' not found.")
        
        
        
        


from yamet.common.utils import grep
from typing import List, Union, Tuple, Any

class CannotGetPropertyException(Exception):
    """Raised when a property cannot be obtained from output files.

    Args:
        Exception (str): Property which causes the error
    """
    def __init__(self, prop: str) -> None:
        self.proprty = prop
        self.message = f"Cannot get {prop}."
        super().__init__(self.message)


def get_pressure(include_pulay:bool = False, all_frames: bool = False) -> Union[List[float], float]:
    """Using GREP to get the pressure from an OUTCAR

    Args:
        include_pulay (bool, optional): Defaults to False.
        all_frames (bool, optional): Defaults to False to use the last one

    Raises:
        CannotGetPropertyException: Raised if pressure cannot be grepped.

    Returns:
        Union[List[float], float]: List of pressures for all_frames = True; float if False
    """
    lines, sucess = grep("OUTCAR", "pressure")
    if not sucess:
        raise CannotGetPropertyException("pressure")
    for line in lines:
        line = line.decode('UTF-8')
        ext = line.split()[3]
        pul = line.split()[8]
        if include_pulay:
            p = float(ext) + float(pul)
        else:
            p = float(pul)
        pressures.append(p*0.1)
    pressures = []
    if all_frames:
        return pressures
    else:
        return pressures[-1]
    
    


class Binary:
    def __init__(self) -> None:
        pass

class Ternary:
    def __init__(self) -> None:
        pass

class Ternary3D:
    def __init__(self) -> None:
        pass

class Quaternery:
    def __init__(self) -> None:
        pass


import pandas as pd
from typing import Union, List, Tuple, Any
from yamet.chemistry.utils import Composition, UnbalanceableEquationError, balance
#from yamet.hull.vertical_distance import vertical_distance
from scipy.spatial import ConvexHull
import pandas as pd
import numpy as np
from pprint import pprint

class Hull:
    """Convex Hull Class

    desc.

    Attributes:
        hull_data:
            The dataframe containing hull data, referencing the primary table
        y
        
    """
    def __init__(
        self,
        structure_table: pd.DataFrame,
        compositions: Union[List[str], str],
        ignore: Any = None,        
                ) -> None:
        """ Initiliastion for Hull

        Args:
            data:
                The data source to build the hull from
        
        """
        if not isinstance(structure_table, pd.DataFrame):
            raise TypeError("The source data for a hull construction needs to be a dataframe object.")
        self.structure_table = structure_table.reset_index()

        if isinstance(compositions, list):
                self.compositions = [Composition(comp) for comp in compositions]
        elif isinstance(compositions, str):
                self.compositions = [Composition(comp) for comp in compositions.split(":")]

        # Ok, now we are done with the argument handling...
        self.reduced_formulae = [comp.get_formula(reduced = True) for comp in self.compositions]
        self.formulae = [comp.get_formula(reduced = False) for comp in self.compositions]

        self.concentration_interpolation()
        self.get_relative_formation_energies()
        self.make_hull()
        # with pd.option_context('display.max_rows', None,
        #         'display.max_columns', None,
        #         'display.precision', 3,
        #         'display.width', None
        #         ):
        #     print(self.hull_table)

    def concentration_interpolation(self, history = None) -> List[List[int]]:
        """Ascertain the concentration of each datapoint according to the sleceted endpoints.

        Args:

        Returns:
            A nested list of the indices of the endpoint in the parent dataframe, self.structure_table
        """
        def get_concs(ID, formula, endpoint_formulae, error_val = np.nan):
            ID = int(ID)
            if Composition(formula).get_formula(reduced=True) in endpoint_formulae:
                 nfu = Composition(formula).formula_units
                 ratios = [0] * len(endpoint_formulae)
                 ratios[endpoint_formulae.index(Composition(formula).get_formula(reduced=True))] = nfu
                 return ID, *ratios, 1
            try:
                ratios = balance([Composition(endpoint) for endpoint in endpoint_formulae], [Composition(formula)])
            except UnbalanceableEquationError:
                return ID, *[error_val] * (len(endpoint_formulae) + 1)
            sum_r_ratio = sum(ratios[0])
            return ID, *ratios[0], *ratios[1]
        
        endpoints_indices = []  
        missing = []
        for reduced_composition in self.reduced_formulae:
            indices = list(np.where(self.structure_table.formula_reduced == reduced_composition)[0])
            endpoints_indices.append(indices)

            # Doing this so user has a full picture of what's missing in one go, rather than adding something
            # only to find something else is missing.
            if len(indices) == 0:
                missing.append(reduced_composition)
        if any([len(idc) == 0 for idc in endpoints_indices]):
            newline = "\n" # f-strings are not as wonderful as they first appear
            raise Exception(f"\n\nMissing endpoint structure(s):\n{newline.join(missing)}\n")
        
        self.hull_table = pd.DataFrame(data = self.structure_table.apply(lambda x: get_concs(x["index"], x["formula"], self.reduced_formulae), axis=1, result_type="expand"))
        self.hull_table.rename(columns=dict(enumerate(["structIdx"] + self.reduced_formulae + ["prod"])), inplace=True)
        self.hull_table = self.hull_table.astype({"structIdx": int})
        self.hull_table.dropna(how = "any")

        self.hull_table = self.hull_table.assign(
             energy = lambda x: self.structure_table.iloc[x["structIdx"]]["energy"],
             atoms_count = lambda x: self.structure_table.iloc[x["structIdx"]]["atoms_count"],
             formula_units = lambda x: self.structure_table.iloc[x["structIdx"]]["formula_units"],
             pressure = lambda x: self.structure_table.iloc[x["structIdx"]]["pressure"],
        )
        self.hull_comps = self.hull_table.formula_units.tolist()
        self.hull_table.reset_index(inplace=True, drop=True)
        self.endpoints_indices = endpoints_indices

    def get_relative_formation_energies(self):
        """ Add the relative formation energies per atom to the hull dataframe.        """
        def relHf(row):
            refs = reference_fu_energies
            energy = row["energy"]
            r_ratios = [row[name] for name in self.reduced_formulae]
            p_ratio = row["prod"]
            min_fu_sum = 0

            # can't use sum as the lambda function uses eries objects, not single values.
            for ref, rto in zip(refs, r_ratios):
                min_fu_sum = min_fu_sum + (rto * ref)
            return ((energy - min_fu_sum) / row["atoms_count"] / row["prod"])
        
        def get_x(row):
            r_tot = 0
            for form in self.reduced_formulae:
                r_tot += row[form]
            return row[self.reduced_formulae[-1]] / r_tot

        mins = []
        
        indices = []
        reference_fu_energies = []
        for idx, epi in enumerate(self.endpoints_indices):
            table = self.hull_table[self.hull_table["structIdx"].isin(epi)]
            m = min(table["energy"])
            mins.append(m)
            indices.append(table['structIdx'][table["energy"] == m].values[0])
            reference_fu_energies.append(m/self.hull_comps[idx])

        if len(self.reduced_formulae) == 2:
            self.hull_table = self.hull_table.assign(
             x = lambda x: get_x(x)
        )


        self.hull_table = self.hull_table.assign(
             rel_form_energy_per_atom = lambda x: relHf(x)
        )
        
        # Keep the table pretty for the user <3
        self.hull_table.drop(labels = ["atoms_count", "formula_units"], axis = 1, inplace = True)


    def make_hull(self) -> pd.DataFrame:
        
        # Energy must always be on the final axis, no matter how many dimensions.

        sorted_hull = self.hull_table.drop(self.hull_table[self.hull_table.rel_form_energy_per_atom > 0].index)
        
        #cvx = ConvexHull()
        pass

    def in_simplex(simplex, points):
        pass


from typing import List
from typing import Union

import pandas as pd
from dscribe.descriptors import SOAP
from matminer.featurizers.structure import OrbitalFieldMatrix

from yamet.common.utils import *
from yamet.collectors.singles import Data as YametDataObject


def get_soaps(df: pd.DataFrame, species: List[Union[int, str]] = None):
    if species == None:
        species = flatten(df["species"].to_list())
    soap = SOAP(
        species=species,
        r_cut=6.0,
        l_max=8,
        n_max=8,
        periodic=True,
        sparse=False,
        average="inner",
        crossover=True,
        rbf="gto",
        sigma=0.1,
    )
    soaps = soap.create(df["atom_obj"].to_list())


class UMAP:
    def __init__(self) -> None:
        pass



import string
from collections import Counter
from functools import reduce
from math import gcd
from fractions import Fraction
from typing import Any, Dict, List, Literal, Tuple, Union

import chemparse
import numpy as np
from sympy import Matrix, symbols as sympy_symbols
from sympy.solvers.solveset import linsolve
from yamet.common.utils import *

class UnbalanceableEquationError(Exception):
    """Raised when an equation cannot be balanced"""
    pass

class Composition:
    def __init__(self, formula) -> None:
        if isinstance(formula, dict):
            self.composition = formula
        elif isinstance(formula, str):
            self.composition = self._from_string(formula)
        elif isinstance(formula, list):
            self.composition = self._from_list(formula)
        else:
            raise TypeError("Unexcepted type. Must be a string, list, or dictionary.")
        self.reduced, self.formula_units = self._reduce()

    def get_composition(self, reduced: bool = False):
        if reduced:
            return self.reduced
        else:
            return self.composition

    def get_formula(self, reduced: bool = False, style: Union[None, str] = None):
        assert style in [None, "tex", "pretty"]
        if reduced:
            comp = self.reduced
        else:
            comp = self.composition

        tex = style == "tex"
        counts = []
        # For correctness, we should put chalcogen/halogen stuff at the end
        terminations = ["C", "N", "P", "S", "Br", "Cl", "F", "O", "H"]
        for element, count in comp.items():
            counts.append((element, count))
        counts.sort(key=lambda x: x[0])
        elements = list(map(list, zip(*counts)))[0]
        # Moving the termination elements to the end in turn
        for termination in terminations:
            if termination in elements:
                idx = elements.index(termination)
                counts.append(counts.pop(idx))
        list_str = []
        for symbol, count in counts:
            if count > 0:
                list_str.append(symbol)
            if count != 1:
                if float(count).is_integer():
                    count = int(count)
                if tex:
                    list_str.append(f"_{{{count}}}")
                else:
                    list_str.append(count)

        if tex:
            return f"${''.join(list_str)}$"
        elif style == "pretty":
            s = "".join([str(s) for s in list_str])
            sb = s.maketrans("0123456789", "₀₁₂₃₄₅₆₇₈₉")
            return s.translate(sb)
        else:
            return "".join([str(s) for s in list_str])

    def get_atom_count(self) -> int:
        """ Get the total number of atoms in a composition object
    
        Returns (int):
            Number of atoms in a composition
        """
        count = 0
        for k, v in self.composition.items():
            count += v
        return count

    # Type handling functions    
    
    def _from_string(self, formula) -> Dict:
        return chemparse.parse_formula(formula)

    def _from_list(self, formula) -> Dict:
        comp = dict()
        if isinstance(formula[0], str):
            return dict(Counter(formula))
        elif isinstance(formula, (tuple, list)):
            santised = [(s, float(c)) for s, c in formula]
            for symbol, count in santised:
                if symbol not in comp.keys():
                    comp[symbol] = count
                else:
                    comp[symbol] += count
            return comp

    def _reduce(self) -> Dict:
        (elements, counts) = zip(*self.composition.items())
        if not any([float(x).is_integer() for x in counts]):
            print("Cannot reduce non-integer compositions (yet!)")
            return self.composition
        counts = [int(count) for count in counts]
        denom = reduce(gcd, counts)
        rratio = [n / denom for n in counts]
        return dict(zip(elements, rratio)), denom
    

def balance(reactants: List[Composition], products: List[Composition], set_missing_to_zero: bool = True) -> Tuple[List[float], List[float]]:
    """Using linear algebra to solve chemical equation balancing
    
        Args:
            reactants:
                A list of YAMET Compositions for the reactants
                of the equation (LHS)

            product:
                A list of YAMET Compositions for the products
                of the equation (RHS)
                
        Returns:
            A nested tuple of the ratios of the products and reactants.
            i.e ((reactant ratios in order of input), (product ratios in order of input))
    
    """
    def lcm(a, b):
        return a*b // gcd(a, b)
    
    def frac_gcd(*nums):
        fractions = [Fraction(num).limit_denominator() for num in nums]
        mult = reduce(lcm, [frac.denominator for frac in fractions])
        mints = [int(frac*mult) for frac in fractions]
        div = reduce(gcd, mints)
        return [int(n / div) for n in mints]

    r_compositions = [r.get_composition() for r in reactants]
    p_compositions = [p.get_composition() for p in products]

    nve_p_comp = []

    rset = set(flatten([list(r.keys()) for r in r_compositions]))
    pset = set(flatten([list(p.keys()) for p in p_compositions]))
    if pset != rset:
        print(f"Unable to balance equation:\n reactants:\t{rset}\n products:\t{pset}\n\n")
        raise UnbalanceableEquationError
    

    chem_matrix = np.zeros((len(rset), len(reactants + products)))

    #SymPy symbols...
    chars = [string.ascii_lowercase[i] for i, _ in enumerate(reactants + products)]
    syms = sympy_symbols(f"{' ,'.join(chars)}")

    for p_comp in p_compositions:
        npc = p_comp
        for k, v in p_comp.items():
            npc[k] = -v
        nve_p_comp.append(npc)

    compositions = r_compositions + nve_p_comp
    for idx, comp in enumerate(compositions):
        for jdx, elem in enumerate(list(rset)):
            chem_matrix[jdx][idx] = comp.get(elem, 0.0)
    
    A = Matrix(chem_matrix)
    b = Matrix(np.zeros((np.shape(A)[0], 1)))

    solution = linsolve((A, b), syms)
    sols = []
    for sol in list(solution)[0]:
        sols.append(sol.subs(syms[-1], 1))
    try:
        amounts = frac_gcd(*[str(s) for s in sols])
    except ZeroDivisionError:
        raise UnbalanceableEquationError
    return [amounts[:len(r_compositions)], amounts[-len(p_compositions):]] 

# General python helper shenanigans
def cprint(s, colour = None):
    if colour == None:
        print(s)

def stop():
    cprint("Exitting.")
    exit()


from typing import List, Tuple, Any
from contextlib import contextmanager
import os
import subprocess as sp
from yamet.analysis.symmetry import get_spacegroup_info
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.core import Structure
from ase.io.res import Res
from ase import Atoms


def flatten(l: List[List[Any]]) -> List[Any]:
    return [item for sublist in l for item in sublist]


def tail(f: str, lines: int = 20) -> List[str]:
    """Tailing with python!

    Args:
        f (str): Path to file to be tailed
        lines (int, optional): Number of lines. Defaults to 20.

    Returns:
        (List[str]): The last $n$ lines of file f.
    """
    p = sp.Popen(["tail", f"-n{lines}", f], stdout=sp.PIPE)
    return [line.strip() for line in p.stdout.readlines()]


def grep(f: str, query: str) -> Tuple[List[str], bool]:
    """Grepping with python!

    Args:
        f (str): The path to the file to be grepped
        query (str): The substring to be searched for

    Returns:
        Tuple[List[str], bool]: The lines coutaining the found query, and
            whether or not the search was successful
    """
    p = sp.Popen(["grep", query, f], stdout=sp.PIPE)
    lines = [line.strip() for line in p.stdout.readlines()]
    if len(lines) > 0:
        return lines, True
    else:
        return lines, False


@contextmanager
def cd(newdir):
    """Move to and operate in new dirctory.

    Best used in with statements, i.e:

    ```
        with cd("my_dir"):
            # Doing things in my_dir
        # Now we're back in the parent directory
    ```

    Args:
        newdir (_type_): _description_
    """
    prevdir = os.getcwd()
    npath = f"{prevdir}/{newdir}"
    if not os.path.exists(npath):
        os.mkdir(newdir)
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def atoms2res(
    atoms: Atoms,
    fname: str,
    energy: float = None,
    pressure: float = None,
    spacegroup: str = None,
):
    if energy == None:
        energy = atoms.get_potential_energy()
    if pressure == None:
        pressure = atoms.get_pressure()
    if spacegroup == None:
        struct = atoms2struct(atoms)
        spacegroup = get_spacegroup_info(struct)[0]

    res = Res(
        atoms, pressure=pressure, energy=energy, spacegroup=spacegroup, times_found=1
    )
    res.write_file(fname)


def atoms2struct(atoms) -> Structure:
    return AseAtomsAdaptor(atoms).get_struct()


def struct2atoms(struct) -> Atoms:
    return AseAtomsAdaptor(struct).get_atoms()



from pprint import pprint

def handle_magmoms(atoms, magmoms = None):
    """
    The magamoms variable should be a list parsed in the form:
    [...] -mgm Fe 5 Nb 0.6 O 0.6 [...]
    which is then converted to a dictionary:
    d = {
        'Fe': 5.,
        'Nb': 0.6,
        'O': 0.6
        }
        
        
    """
    if magmoms is None:
        return atoms
    else:    
        elements = magmoms[::2]
        values = magmoms[1::2]
        d = dict(zip(elements, values))
        init_mgm = []
        for s in atoms.get_chemical_symbols():
            if s not in elements:
                init_mgm.append(0)
            else:
                init_mgm.append(d[s])
        atoms.set_initial_magnetic_moments(init_mgm)
        return atoms

def handle_hubbard(atoms, luj):
    """Generate a dictionary of Hubbard parameters to be applied with ASE calculators

    Args:
        atoms (ase.Atoms): The ASE atoms object
        luj (list): A list of parameters in the form
        
        [<symbol>, <l_number>, <U_correction>, <J_corrrection>]
        
        This is so one can easily use this with nargs in argparse.

    Returns:
        dict: Dictionary to be applied to 
    """
    if luj is None:
        print("Hubbard corrections not set.")
        return None
    labels = ["L", "U", "J"]
    n = 4
    elements = []
    d = {}
    separated = [luj[i : i + n] for i in range(0, len(luj), n)]
    for indiv in separated:
        elements.append(indiv[0])
        d[indiv[0]] = dict(zip(labels, [float(x) for x in indiv[-3:]]))
    for s in atoms.symbols:
        if s not in elements:
            d[s] = dict(zip(labels, [-1, 0, 0])) # Negative 1 for no onsite interation added.
    pprint(d)
    return d


import shelve
import traceback
import warnings
from copy import copy
from copy import deepcopy
from glob import glob
from typing import Any
from typing import Callable
from typing import List
from typing import Tuple
from typing import Union
from xml.etree.ElementTree import ParseError as XMLParseError

import numpy as np
import pandas as pd
from ase.io import read as ase_read
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.io.vasp.outputs import Vasprun
from tqdm import tqdm

from yamet.analysis.symmetry import get_spacegroup_info
from yamet.chemistry.utils import Composition
from yamet.common.utils import *


warnings.filterwarnings("ignore")

# TODO
# [] Check for convergence when loading in a actual output file, not just a geometry file etc.


class Data:
    """The object used by YAMET to access, analyse, and alter provided structure information.

    Essentially, this is a fancy wrapper for a pandas dataframe object but with some additional
    bits and pieces stuck on the side. That is not to say it's inhereited from pd.DataFrame().
    Rather, the data is stored in a dataframe produced within the class.

    Attributes:
        name: a name for the data collection.
        description: a description of the data collection.
        data: the pd.DataFrame() onject with, you guessed it, the data.
        columns: the columns for the pd.DataFrame which can be appended to at initialisation.

    """

    def __init__(
        self,
        name: str = "",
        description: str = "",
        cache: bool = True,
        cache_id: str = "ALL_DATA",
        additional_fields: Union[List[str], None] = None,
        additional_field_functions: Union[List[Tuple[Any, Any]], None] = None,
        no_property_default: Any = None,
    ) -> None:
        """Initialises the Data class


        name:
            a name for the data collection.

        description:
            a description of the data collection.

        additional_fields:
            A list of strings naming custom columns to add to the dataframe at the initial import phase.

        additional_field_functions:
            A list of tuples of functions and arguments that return the information desired for the
            user defined field. For example:

            [(func_one, args), (func_two, args)]

            The first argument of the user-defined function is assumed to be the file with the
            structure information, such as a .cif, .res etc., so define your
            function as follows:

            def my_function(structure_file, foo, bar):
                # Import the file with your favourite tool
                # Do things with foo, bar and the stuff from the structure file.
                return result # Which will be added to the dataframe in the matching column name in additional_fields

        no_property_default:
                The value to default to if a property cannot be found. Currently used when trying to get energy, stress, and pressures only.
                Others properties are considered too critical so should always be present.

        """
        self.name = name
        self.description = description
        self.sources = []
        self.data = pd.DataFrame()
        self.no_prop_override = no_property_default
        self.cache = cache
        init_columns = [
            "source",
            "atoms_obj",
            "a",
            "b",
            "c",
            "alpha",
            "beta",
            "gamma",
            "cell",
            "formula",
            "formula_reduced",
            "formula_units",
            "species",
            "atoms_count",
            "positions_cart",
            "positions_frac",
            "energy",
            "energy_per_atom",
            "energy_per_formula_unit",
            "pressure",
            "stress",
            "spacegroup",
            "spacegroup_number",
            "crystal_system",
            "converged_ionic",
            "label",
        ]
        if additional_fields is not None:
            assert len(additional_fields) == len(additional_field_functions)
            self.columns = tuple(init_columns + additional_fields)
            self.additional_field_functions = tuple(additional_field_functions)
        else:
            self.additional_field_functions = []
            self.columns = init_columns
            self.no_prop_override = np.nan

        self.cache_id = cache_id
        if self.cache:
            self.from_cache()

    def load(
        self,
        files_location: str,
        label: Union[int, float, str] = ".",
        ignore_unreadable: bool = False,
    ) -> pd.DataFrame:
        """Imports listed files into a dataframe.

        Args:
            files_locations:
                A string or list of strings that can be queried, for example:

                'my_directory/*.cif'
                or
                ['directory1/*.res', directory2/*/vasprun.xml]

            label:
                A user-defined label or tag for the dataset being loaded to
                distinguish it from other data in the frame.

            ignore_unreadable:
                Ignore structures that have properties that cannot be read, such
                as NoneType energies or pressures.

            Returns:
                A pandas dataframe with each row assigned to an imported structure
        """

        df = pd.DataFrame(columns=self.columns)
        files = glob(files_location)
        self.sources += files

        # There must be some way to speed this up?? Parallelisation?
        for idx, f in tqdm(enumerate(files), total=len(files)):
            atoms = ase_read(f)
            struct = AseAtomsAdaptor.get_structure(atoms)
            symbols = atoms.get_chemical_symbols()
            comp = Composition(symbols)
            cell = np.array(atoms.get_cell())
            energy = self._check(atoms.get_potential_energy)
            pressure = self._check(atoms.info.get, ["pressure"])
            stress = self._check(atoms.get_stress)
            additionals = [
                func(f, *args) for func, args in self.additional_field_functions
            ]
            df.loc[idx] = [
                f,
                atoms,
                *atoms.cell.cellpar(),
                cell,
                comp.get_formula(),
                comp.get_formula(reduced=True),
                comp.formula_units,
                tuple(symbols),
                len(symbols),
                atoms.get_positions(),
                atoms.get_scaled_positions(),
                energy,
                energy / len(symbols),
                energy / comp.formula_units,
                pressure,
                stress,
                *get_spacegroup_info(struct),
                "Unknown",
                label,
            ] + additionals
        self.data = pd.concat([self.data, df])
        if ignore_unreadable:
            self.data.dropna(inplace=True)
        self.data = self.data.reset_index(drop=True)
        if self.cache:
            self.to_cache()

    def to_cache(self, overwrite=True, clean=True):
        if clean:
            self.clean()
        with shelve.open("yamet.cache", writeback=True) as cache:
            if self.cache_id == "":
                self.cache_id = f"ALL_DATA"
                print(f"Using {self.cache_id} as cache ID")
            present_keys = cache.keys()
            if self.cache_id in present_keys and not overwrite:
                print(f"Cannot overwrite cache entry {self.cache_id}")
            else:
                cache[self.cache_id] = self.data

    def insert_data(self, df: pd.DataFrame, clean=True):
        bkp = deepcopy(self.data)
        try:
            self.df = pd.concat([self.data, df])
        except:
            print("\nOh no!\n")
            print(traceback.format_exc() + "\n")
            self.df = bkp
            del bkp

    def from_cache(self):
        with shelve.open("yamet.cache") as cache:
            if self.cache_id in cache:
                self.data = cache[self.cache_id]
            else:
                print(
                    f"Cannot find cache entry with ID: {self.cache_id}.\nNothing loaded."
                )

    def clean(self, ignore_fields: list = [], verbose=True):
        """A santisation function for the Data.data data.

        Args:
            A list of fields to ignore when looking for duplicates.
        """

        # Ignoring uncomparable fields
        ignore_fields += [
            "cell",
            "composition",
            "positions_cart",
            "positions_frac",
            "label",
            "converged_ionic",
        ]
        ignore_fields = list(set(ignore_fields))
        subset = [field for field in self.columns if field not in ignore_fields]
        original_length = len(self.data.index)

        # We need this hack to remove the non-hashable arrays
        self.data["cell_tuple"] = self.data.cell.apply(lambda x: tuple(x.flatten()))
        self.data["posfrac_tuple"] = self.data.positions_frac.apply(
            lambda x: tuple(x.flatten())
        )

        tupled = ["cell_tuple", "posfrac_tuple"]
        subset += tupled

        duplicates = self.data.duplicated(subset=subset)
        self.data = self.data[~duplicates]
        num_removed = original_length - len(self.data.index)
        if verbose:
            print(f"{int(num_removed)} duplicates found and removed.")
        for tup in tupled:
            self.data.drop(tup, axis=1, inplace=True)
        self.data.reset_index(inplace=True, drop=True)

    def get_converged(self):
        def from_vasprun(src):
            try:
                v = Vasprun(src)
                return v.converged_ionic
            except XMLParseError:
                return "Unknown"

        self.data["converged_ionic"] = self.data.source.apply(lambda x: from_vasprun(x))

    def _check(self, x, args=[]) -> Any:
        """Catch and return a default if a property cannot be found."""
        try:
            prop = x(*args)
        except:
            prop = self.no_prop_override
        return prop

    def get_data(self):
        return copy(self.data)
